Module Inner
    
! 01/06/2021 : Variable explanations and a continuous version of simulation are attached.
    
! innerLoop.f90 is a module for subroutines to compute values, distributions and simulation results
! Endgrid : subroutine to solve values
! simulation : subroutine for simulations
! distribution : subroutine to compute the distribution
! bellmanerror : subroutine to compute Bellman errors
    
    use parameters
    use comp_libs
    use omp_lib
    
implicit none

    contains
    
Subroutine Endgrid(zerocap,proy,endow,bgrid,bprice,gb,gc,gd,Vd,Mval)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! subroutine to solve values !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    
Implicit none


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! input !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! zerocap : index of 0 bond on bond grid (bgrid)
! proy : markov chain for endowment
! endow : endowment
! bgrid : grid for bond
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! output !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! bprice : bond price function
! gb : decision rule for bond
! gc : decision rule for consumption
! gd : decision rule for default
! Vd : value fucntion for default
! Mval : max(Vd,Vr) where Vr is repaying value function
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

Integer, intent(in) :: zerocap
Real(8), intent(in) :: proy(Ny,Ny), endow(Ny),bgrid(Nb)
Real(8), intent(out) :: bprice(Nb,Ny),gb(Nb,Ny)
Real(8), intent(out) :: gc(Nb,Ny),gd(Nb,Ny)
Real(8), intent(out) :: Vd(Ny),Mval(Nb,Ny)
Real(8) :: tvgr(Nb,Ny), tvd2(Ny), tvd1(Ny), Edval(Ny)
Real(8) :: inival(Nb,Ny), inivd(Ny)
Real(8) :: gbgtemp(Nb,Ny),gcgtemp(Nb,Ny)
Real(8) :: tval1(Nb,Ny), tval2(Nb,Ny), Eval(Nb,Ny), Etval(Nb,Ny)
Real(8) :: tempmat1(Nb,Ny), tempmat2(Nb,Ny), tbp1(Nb,Ny),tbp2(Nb,Ny)
Real(8) :: pry(Ny), DEv(Nb,Ny),Dbprice(Nb,Ny), Dqb(Nb,Ny),qb(Nb,Ny)
Real(8) :: DDqb(Nb,Ny),Coef(4,bnotsb+1)
Real(8) :: Coef1(4,bnotsb+1),Coef2(4,bnotsb+1),Coef3(4,bnotsb+1)
Real(8) :: valb1(Nb),valb2(Nb),valb3(Nb),valb4(Nb),expDval(Nb),envelcons(Nb,Ny),DqbDEv(Nb,Ny)
Real(8) :: Coefy1(4,ynotsy+1), coefy2(4,ynotsy+1), valy1(Ny), valy2(Ny)
Real(8) :: conendg(Nb,Ny), cohendg(Nb,Ny), bendg(Nb,Ny)
Real(8) :: valtempg(Nb,Ny),btend(Nb), cutoffmat(ny), cutoff
Real(8) :: cons, autinc, edwinc,util,vep1,vep2,vep3,vep4,val1,val2, val3, val4
Real(8) :: aaa,bbb,ccc,ddd,bp,coh,bval,bpval,wb, ycutoff, ythres(Nb)
Real(8) :: diffprice, diffval, diff, mindval,cohp,start,finish

Real(8) :: bpair1(nb), bpair2(nb), wbp, wgt, bpgrid(ny)


Integer :: ind_b1, ind_b2, ind_y1, ind_y2, i1, i2, j1, j2, k1, k2, iter, iterp,tbnum
Integer :: rblim(Ny), indminb(Ny), indmaxb(Ny),blowbdd, cblowbdd, bini, blow, yind, bplow
Integer :: imin, imax, gloind(Nb,Ny), subind(Nb),bsol,bupsol,bhigh,gridbnum,innerwrite

Integer :: inigloind(ny)

Logical :: mask1(Nb), mask2(Nb), endo(Nb)
!!print*, endow

call cpu_time(start)




!******************************************* Arrange the initial setting ****************************************************************************

!!!!!!!!!!!!!!!!!!! Set-up initial values functions, tvd and tvgr !!!!!!!!!!!!!!!!!!!!!!!!!!!
Do ind_y1=1,Ny
    edwinc=endow(ind_y1)
    Do ind_b1=1,Nb
        cons=edwinc+bgrid(ind_b1)
        If (cons>0.0d0) then
            tval2(ind_b1,ind_y1)=util(risksig,cons)
        else
            tval2(ind_b1,ind_y1)=-1d+8
        end if
    end do
    tvd2(ind_y1)=util(risksig,edwinc)
end do
!tvd2=-1d+8
tbp2 = 1.0d0/rf
tbp2(1:zerocap-1,:) = 0.0d0
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Check the differences between the updated ones (value unfunction and loan price function) and the previou ones) !!!!!!!!!!!!!!!!!!!!!!!!!!!!

diff=eps*1d+3
iter=0

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!*********************************************************  End of the arrangement for the itial setting ****************************************************************

!!!!!!!!!!!!!!! if you want to check the outcomes, innerwrite=1. Otherwise innerwrite=0.
innerwrite=0
    
!**********************************************************  Start of the value function Iteration *************************************************************************
Do while(diff>eps)
    
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <Begin with Step 9> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!! Update the price and value functions !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    tval1=tval2
    tvd1=tvd2
    tbp1=tbp2
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Calculate the size of loans q x b !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    Do ind_y1=1,Ny
        qb(:,ind_y1)=tbp1(:,ind_y1)*bgrid
    end do
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <End of Step 9> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <Begin with Step 10> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!! Calculate the expected values, Eval and Edval. !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    Do ind_y1=1,Ny
        pry=proy(ind_y1,:)
        valb1= 0.0d0
        vep1=0.0d0
        Do ind_y2=1,Ny
            valb1=valb1+beta*pry(ind_y2)*tval1(:,ind_y2)
            vep1=vep1+beta*pry(ind_y2)*tvd1(ind_y2)
        end do
        Eval(:,ind_y1)=valb1
        Edval(ind_y1) = vep1
    end do
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <End Step 10> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <Begin with Step 8> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!! Calculate the defaulting value tvd2 !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    Do ind_y1=1,Ny
        edwinc=endow(ind_y1)
        cons=autinc(ybar,edwinc)
        tvd2(ind_y1)=util(risksig,cons)+theta*Eval(zerocap,ind_y1)+(1.0d0-theta)*Edval(ind_y1)
    end do
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <End with Step 10> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! soltype=2 => grid search !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    If (soltype==2) then
        !!$omp parallel default(shared) private(ind_y1,ind_b1,valb1,ind_b2,cons,bsol) num_threads(39)
        Do ind_y1=1,Ny
            Do ind_b1=1,Nb
                valb1=-1d+8
                Do ind_b2=1,Nb
                    cons=endow(ind_y1)+bgrid(ind_b1)-qb(ind_b2,ind_y1)
                    If (cons>0.) then
                        valb1(ind_b2)=util(risksig,cons)+Eval(ind_b2,ind_y1)
                    else
                        valb1(ind_b2)=-1e+8
                    end if
                end do
                bsol=maxloc(valb1,1)
                tvgr(ind_b1,ind_y1)=valb1(bsol)
                !print*, valb1(bsol)
                gcgtemp(ind_b1,ind_y1)=endow(ind_y1)+bgrid(ind_b1)-qb(bsol,ind_y1)
                gbgtemp(ind_b1,ind_y1)=bgrid(bsol)
            end do
        end do
        !!$omp end parallel

    else if (soltype==1) then!!!!! soltype1=1 => endogenous grid method !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <Begin with Step 1> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Calculate the derivative of q(b',y) b' !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        valb1=0.
        valb2=0.
        Do ind_y1=1,Ny
            valb1=tbp1(:,ind_y1)
            call numderiv(bnotsb,bgrid,valb1,valb2,deriv)
            Dqb(:,ind_y1)=valb2*bgrid+valb1
        end do
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Calculate the risky-borrowing limits !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        valb1=0.
        Do ind_y1=1,Ny
            valb1=Dqb(:,ind_y1)
            dqloop1: Do ind_b2=1,Nb
                vep1=minval(valb1(ind_b2:nb))
                blowbdd=ind_b2
                If (vep1>0.0d0) exit dqloop1
            end do dqloop1

            rblim(ind_y1)=blowbdd
            !print*, blowbdd,ind_y1
        end do
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Calculate the derivative of the expected value function DEv(b'y) and q(B',y)B' !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        valb1=0.
        valb2=0.
        Do ind_y1=1,Ny
            valb1=Eval(:,ind_y1)
            call numderiv(bnotsb,bgrid,valb1,valb2,deriv)
            DEv(:,ind_y1)=valb2
        end do
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        !print*, 'here2'
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Calculate (1/Dqb)*Edv=DqbDEv !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        Do ind_y1=1,Ny
            blowbdd=rblim(ind_y1)
            DqbDEv(blowbdd:nb,ind_y1)=Dev(blowbdd:nb,ind_y1)/Dqb(blowbdd:nb,ind_y1)
        end do
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <End of Step 1> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <Begin with Step 2> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Divide concave and non-concave regions for b'!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        indminb=Nb
        indmaxb=rblim
        Do ind_y1=1,Ny
            valb1=DqbDEv(:,ind_y1)
            blowbdd=rblim(ind_y1)
            mask1(blowbdd:Nb-1)=valb1(blowbdd:Nb-1) - valb1(blowbdd+1:Nb) < 0.0d0
            mask2(blowbdd+1:Nb)=mask1(blowbdd:Nb-1)
            If (count(mask1(blowbdd:Nb-1))>0) then !!!! non-concavity exists!!!!
                val1=maxval(valb1(blowbdd+1:Nb),mask2(blowbdd+1:Nb)==.true.)
                val2=minval(valb1(blowbdd:Nb-1),mask1(blowbdd:Nb-1)==.true.)
                !print*, val1, val2, ind_y1
                If (count(valb1(blowbdd:)>val1)==0) then
                    indminb(ind_y1)=blowbdd
                else
                    indminb(ind_y1)=minloc(valb1(blowbdd:),1,valb1(blowbdd:)>val1)+(blowbdd-1)+1
                end if
                If (count(valb1(blowbdd:)<val2)==0) then
                    indmaxb(ind_y1)=Nb
                else
                    indmaxb(ind_y1)=maxloc(valb1(blowbdd:),1,valb1(blowbdd:)<val2)+(blowbdd-1)-1
                end if
                !print*, count(mask(blowbdd+1:)),count(expDval(blowbdd:)>val1),count(expDval(blowbdd:)<val2)
                !print*, indminb(ind_y1),indmaxb(ind_y1), ind_y1, val1, val2
            end if
        end do
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <End of Step 2> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        !print*, 'here4'
        
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <Begin with Step 3 & 4> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Use F.O.C with the endogenous grid method on the whole region of B' !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Save the implied consumption, cash on hand, and current bond B. !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        Do ind_y1=1,Ny
            blowbdd=rblim(ind_y1)
            Do ind_b2=blowbdd,Nb
                aaa=DqbDEv(ind_b2,ind_y1)
                cons=aaa**(-1.0d0/risksig)
                cohendg(ind_b2,ind_y1)=cons+qb(ind_b2,ind_y1)
                bendg(ind_b2,ind_y1)=cohendg(ind_b2,ind_y1)-endow(ind_y1)
                If (cons>0.0d0) then
                    valtempg(ind_b2,ind_y1)=util(risksig,cons)+Eval(ind_b2,ind_y1)
                else
                    valtempg(ind_b2,ind_y1)=-1d+8
                end if
            end do
        end do
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <End of Step 3 & 4> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



        !print*, 'here5'

        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <Begin with Step 5> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Save global solutions with non-concave area by checking whether !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! the solutions implied by FOC are global solutions !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        gloind=0
        Do ind_y1=1,Ny
            imin=indminb(ind_y1)
            imax=indmaxb(ind_y1)
            !print*, imin, imax, ind_y1
            blowbdd=rblim(ind_y1)
            bhigh=nb
            Do ind_b2=Nb,blowbdd,-1

                If ((ind_b2<imin).or.(ind_b2>imax)) then
                    gloind(ind_b2,ind_y1)=1
                    !print*, ind_b2, ind_y1, imin, imax
                else
                    coh=cohendg(ind_b2,ind_y1)
                    cons=coh-qb(ind_b2,ind_y1)
                    If (cons>0.0d0) then
                        vep1=util(risksig,cons)+Eval(ind_b2,ind_y1)
                    else
                        vep1=-1d+8
                    end if
                    bsol=ind_b2

                    Do i1=imin,bhigh
                        cons=coh-qb(i1,ind_y1)
                        If (cons>0.0d0) then
                            vep2=util(risksig,cons)+Eval(i1,ind_y1)
                        else
                            vep2=-1d+8
                        end if
                        If(vep1<vep2) then
                            bsol=i1
                            exit
                        end if
                    end do

                    If ((bsol==ind_b2).and.(vep1>-1d+8)) then
                        gloind(ind_b2,ind_y1)=1
                        bhigh=bsol
                    end if

                end if
            end do
            !print*, gloind(:,ind_y1)
        end do
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <End of Step 5> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        !!!!!!!!!!!!!!! Compute the threshold b'=0 on b (not used now)
        !
        !Do ind_y1=1,ny
        !    loop3: Do ind_b2=1,nb
        !        i1 = gloind(ind_b2,ind_y1)
        !        If (i1 == 1) then
        !            inigloind(ind_y1)=ind_b2
        !            exit loop3
        !        end if
        !    end do loop3
        !    bini=inigloind(ind_y1)
        !    blow = bini
        !    !blow = max(zerocap,bini)
        !    cutoffmat(ind_y1)=bendg(blow,ind_y1)
        !end do
        
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <Begin with Step 6> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
         Do ind_y1=1,Ny
            blowbdd=rblim(ind_y1)
            Do ind_b2=blowbdd,Nb                                                
                bendg(ind_b2,ind_y1)=cohendg(ind_b2,ind_y1)-endow(ind_y1)                
            end do
         end do
         !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <End of Step 6> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        
        !print*, 'here7'


        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <Begin with Step 7> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!! Using the index of global solutions, retrieve the value fucntions for good credit status!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        tvgr=-1d+8
        Do ind_y1=1, Ny
            blowbdd=rblim(ind_y1)
            mask1(blowbdd:nb)=gloind(blowbdd:nb,ind_y1).eq.1
            gridbnum=count(mask1(blowbdd:nb))
            bpair2(1:gridbnum) = pack(bgrid(blowbdd:nb),mask1(blowbdd:nb))
            bpair1(1:gridbnum) = pack(bendg(blowbdd:nb,ind_y1),mask1(blowbdd:nb))
            cutoff = cutoffmat(ind_y1)
    

            j1 = 0
            lp2: Do ind_b1=nb, 1, -1
                bval=bgrid(ind_b1)
                If (bval >= cutoff) then
                    call indandwgt_rst(bval,1,gridbnum,1,nb,bpair1,blow,wb)
                    bpval = wb*bpair2(blow) + (1.0d0-wb)*bpair2(blow+1)
                    call indandwgt_rst(bpval,1,nb,1,nb,bgrid,bplow,wbp)
                    bp=wbp*tbp1(bplow,ind_y1)+(1.0d0-wbp)*tbp1(bplow+1,ind_y1)
                    cons = endow(ind_y1) + bval - bp*bpval
                    If (cons > 0.0d0) then
                        gbgtemp(ind_b1,ind_y1) = bpval
                        gcgtemp(ind_b1,ind_y1) = cons
                        tvgr(ind_b1,ind_y1) = util(risksig,cons)+(wbp*Eval(bplow,ind_y1)+(1.0d0-wbp)*Eval(bplow+1,ind_y1))
                    else
                        gbgtemp(ind_b1,ind_y1) = bgrid(1)
                        gcgtemp(ind_b1,ind_y1) = -1d+8
                        tvgr(ind_b1,ind_y1) = -1d+8
                    end if
                else if (bval < cutoff) then
                    valb1=-1d+10
                    valb2=-1d+10
                    j1 = j1 + 1
                    If (j1 == 1) then
                        bpval = gbgtemp(ind_b1+1,ind_y1)
                        call indandwgt_rst(bpval,1,nb,1,nb,bgrid,bplow,wbp)
                        bp=wbp*tbp1(bplow,ind_y1)+(1.0d0-wbp)*tbp1(bplow+1,ind_y1)
                        cons = endow(ind_y1) + bval - bp*bpval
                        bhigh = bplow
                        If (cons > 0.0d0) then
                            valb1(bhigh+1) = util(risksig,cons)+(wbp*Eval(bplow,ind_y1)+(1.0d0-wbp)*Eval(bplow+1,ind_y1))
                            valb2(bhigh+1) = cons
                        else
                            valb1(bhigh+1) = -1d+8
                            valb2(bhigh+1) = -1d+8
                        end if
                    end if

                    Do ind_b2 = bhigh, blowbdd, -1
                        bp = tbp1(ind_b2,ind_y1)
                        cons = endow(ind_y1) + bval - bp*bgrid(ind_b2)
                        If (cons > 0.0d0) then
                            valb1(ind_b2) = util(risksig,cons)+Eval(ind_b2,ind_y1)
                            valb2(ind_b2) = cons
                        else
                            valb1(bhigh+1) = -1d+8
                            valb2(bhigh+1) = -1d+8
                        end if
                    end do

                    If (j1 == 1) then
                        call gridsearch(nb,valb1,blowbdd,bhigh+1,bsol,vep1)
                    else
                        call gridsearch(nb,valb1,blowbdd,bhigh,bsol,vep1)
                    end if


                    tvgr(ind_b1,ind_y1) = valb1(bsol)
                    gcgtemp(ind_b1,ind_y1) = valb2(bsol)

                    If ((bsol == bhigh+1).and.(j1==1)) then
                        gbgtemp(ind_b1,ind_y1) = gbgtemp(ind_b1+1,ind_y1)
                    else
                        gbgtemp(ind_b1,ind_y1) = bgrid(bsol)
                    end if

                    If ((j1==1).and.(bsol==bhigh+1)) then
                        j1 = 0
                    else
                        j1 = 1
                        bhigh = bsol
                    end if

                    If (tvgr(ind_b1,ind_y1) == -1d+8) then
                        tvgr(1:ind_b1-1,ind_y1) = -1d+8
                        gbgtemp(1:ind_b1-1,ind_y1) = bgrid(1)
                        gcgtemp(1:ind_b1-1,ind_y1) = -1d+8
                        exit lp2
                    end if
                end if
            end do lp2
        end do
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <End of Step 7> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        

    end if

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <Begin with Step 9> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Update Tv=max(Vgr,Vd)!!!!!!!!!!!!!!!!!!!
    Do ind_y1=1,Ny
        vep1=tvd2(ind_y1)
        Do ind_b1=1,Nb
            vep2=tvgr(ind_b1,ind_y1)
            !print*, vep1, vep2
            If (vep2>vep1) then
                tval2(ind_b1,ind_y1)=vep2
                gb(ind_b1,ind_y1)=gbgtemp(ind_b1,ind_y1)
                gc(ind_b1,ind_y1)=gcgtemp(ind_b1,ind_y1)
                gd(ind_b1,ind_y1)=0.0d0
                !print*, tval2(ind_b1,ind_y1),gb(ind_b1,ind_y1),gc(ind_b1,ind_y1),gd(ind_b1,ind_y1)
            else
                edwinc=endow(ind_y1)
                cons=autinc(ybar,edwinc)
                tval2(ind_b1,ind_y1)=vep1
                gb(ind_b1,ind_y1)=0.0d0
                gc(ind_b1,ind_y1)=cons
                gd(ind_b1,ind_y1)=1.0d0
            end if
        end do
    end do
    

    
    
    !!!!!!!!!!!!!!!!!!! Upate new prices with differentiability !!!!!!!!!!!!!!!!!!
    
    If (pricetype == 1 ) then ! continuous updating rule
        Do ind_y1=1,ny-1
            wgt= 0.50d0*(log(endow(ind_y1+1)/expy)-log(endow(ind_y1)/expy))
            bpgrid(ind_y1)=log(endow(ind_y1)/expy)+wgt
        end do
        bpgrid(ny)=log(endow(ny)/expy)


        tbp2=dble(1)/rf
        Do ind_b2=1,zerocap-1
            bval=bgrid(ind_b2)
            ycutoff=endow(1)/expy
            !If (bval<0.) then
            !    print*, tvd2(1), tvgr(ind_b2,1)
            !end if
            If (tvd2(1)<tvgr(ind_b2,1)) then
                tbp2(ind_b2,:)=1.0d0/rf
            else if (tvd2(Ny)>=tvgr(ind_b2,Ny)) then
                tbp2(ind_b2,:)=0.0d0
            else
                valy1=tvd2-tvgr(ind_b2,:)
                ybarloop: Do ind_y2=1,Ny
                    val1=tvd2(ind_y2)
                    val2=tvgr(ind_b2,ind_y2)
                    yind=ind_y2
                    If (val2>val1) exit ybarloop
                end do ybarloop
                !val1=valy1(yind-1)
                !aaa=endow(yind-1)
                !val3=valy1(yind)
                !ccc=endow(yind)

                ! 09/25/2020 : take into account the way of tauchen method
                val1=valy1(yind-1)
                aaa=bpgrid(yind-1)
                val3=valy1(yind)
                ccc=bpgrid(yind)



                Do while (abs(aaa-ccc)>eps*1d-1)
                    bbb=0.50d0*(aaa+ccc)
                    !call Blintp(bbb,ynotsy,endow,valy1,val2)
                    ! 09/25/2020 : take into account the way of tauchen method
                    call Blintp(bbb,ynotsy,bpgrid,valy1,val2)
                    !call CubicvalD1(bbb,ynotsy,endow,valy1,coefy1,val2)
                    If (val1*val2>=0) then
                        aaa=bbb
                        val1=val2
                    else
                        ccc=bbb
                        val3=val2
                    end if
                end do
                ! 09/25/2020 : take into account the way of tauchen method
                !ycutoff=bbb/expy
                ycutoff = bbb
                Do ind_y1=1,Ny
                    aaa=ycutoff-rhoy*log(endow(ind_y1)/expy)
                    call normal_cdf(aaa,0.0d0,sigy,vep1)
                    vep2=max(min(1.0d0,vep1),0.0d0)
                    tbp2(ind_b2,ind_y1)=(1.0d0-vep2)/rf
                end do
            end if
        end do
        !print*, 'stupid'


    else if (pricetype == 2) then ! discrete updating rule
        !!!!!!!!!!!!!!!!!!!! Update new prices !!!!!!!!!!!!!!!!!!!!
        tbp2=1.0d0/rf
        Do ind_y1=1,Ny
            pry=proy(ind_y1,:)
            Do ind_b2=1,zerocap-1
                vep1=0.0d0
                Do ind_y2=1,Ny
                    vep1=vep1+pry(ind_y2)*gd(ind_b2,ind_y2)
                end do
                tbp2(ind_b2,ind_y1)=(1.0d0-vep1)/rf
                !print*, vep1
            end do
        end do
    end if

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <End of Step 9> !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !print*, gc(:,1)


    !print*, 'stupid3'
    diffprice=maxval(abs(tbp2-tbp1))
    diffval=maxval(abs(tval1-tval2))
    diff=max(diffval,diffprice)
    !diff=diffval
    iter=iter+1






    If (innerwrite==1) then
        open (unit=19,file="rblim.csv",action="write",status="replace")
        write (19,"(20000(i0.8,',',:))") rblim
        close(19)

        open (unit=19,file="qb.csv",action="write",status="replace")
        Do ind_y1=1,Ny
            write (19,"(20000(f0.8,',',:))") qb(:,ind_y1)
        end do
        close(19)

        open (unit=19,file="ythres.csv",action="write",status="replace")
        write (19,"(20000(f0.8,',',:))") ythres
        close(19)

        open (unit=19,file="Dqb.csv",action="write",status="replace")
        Do ind_y1=1,Ny
            write (19,"(20000(f0.8,',',:))") Dqb(:,ind_y1)
        end do
        close(19)

        open (unit=19,file="tbprice.csv",action="write",status="replace")
        Do ind_y1=1,Ny
            write (19,"(20000(f0.8,',',:))") tbp2(:,ind_y1)
        end do
        close(19)

        open (unit=19,file="DEv.csv",action="write",status="replace")
        Do ind_y1=1,Ny
            write (19,"(20000(f0.8,',',:))") DEv(:,ind_y1)
        end do
        close(19)

        open (unit=19,file="expVal.csv",action="write",status="replace")
        Do ind_y1=1,Ny
            write (19,"(20000(f0.8,',',:))") Eval(rblim(ind_y1):,ind_y1)
        end do
        close(19)

        open (unit=19,file="DqbDEv.csv",action="write",status="replace")
        Do ind_y1=1,Ny
            write (19,"(20000(f0.8,',',:))") DqbDEv(rblim(ind_y1):,ind_y1)
        end do
        close(19)

        open (unit=19,file="endvalue.csv",action="write",status="replace")
        Do ind_y1=1,Ny
            write (19,"(20000(f0.8,',',:))") valtempg(:,ind_y1)
        end do
        close(19)

        open (unit=19,file="bendg.csv",action="write",status="replace")
        Do ind_y1=1,Ny
            write (19,"(20000(f0.8,',',:))") bendg(rblim(ind_y1):,ind_y1)
        end do
        close(19)

        open (unit=19,file="gloind.csv",action="write",status="replace")
        Do ind_y1=1,Ny
            write (19,"(20000(i0.8,',',:))") gloind(:,ind_y1)
        end do
        close(19)

        open (unit=19,file="gb.csv",action="write",status="replace")
        Do ind_y1=1,Ny
            write (19,"(20000(f0.8,',',:))") gb(:,ind_y1)
        end do
        close(19)

        open (unit=19,file="tempvgr.csv",action="write",status="replace")
        Do ind_y1=1,Ny
            write (19,"(20000(f0.8,',',:))") tvgr(:,ind_y1)
        end do
        close(19)
    end if

    if (mod(iter,50) .eq. 0 ) then
    Write(*, '(a)', Advance='no') 'max(dprice,dval) and # of iterations:'
    print*, diff, iter
    !print*, maxloc(abs(tval1-tval2))
    end if




end do

call cpu_time(finish)
print '("Time = ",f10.3," seconds.")',finish-start
    

open (unit=19,file="cputime.csv",action="write",status="replace")   
    write (19,"(20000(f10.3,',',:))") finish-start
    write (19,"(20000(f10.3,',',:))") (finish-start)/dble(iter)
close(19)


bprice=tbp2
Mval=tval2
vd=tvd2

            

End subroutine Endgrid

Subroutine simulation(zerocap,simindex,iidseed,bgrid,endow,bprice,qb,Dqb,gb,gc,gd,simy,simb,simprice,simqb,simcons,simdef,simcredit)
                               
Implicit none
Integer,intent(in) :: zerocap,simindex(Nsim)
Real(8), intent(in) :: iidseed(Nsim),bprice(Nb,Ny),Dqb(Nb,Ny),qb(Nb,Ny)
Real(8), intent(in) :: bgrid(Nb),endow(Ny),gb(Nb,Ny), gc(Nb,Ny),gd(Nb,Ny)
Real(8), intent(out) :: simy(Nsim),simb(Nsim),simprice(Nsim),simqb(Nsim), simcons(Nsim), simdef(Nsim)
Integer, intent(out) :: simcredit(Nsim)
Real(8) :: wb,bval,aaa,bbb,bpval,bvalue,wbp,coef1(4,bnotsb+1),coef2(4,bnotsb+1),coef3(4,bnotsb+1)
Real(8) :: blowval
Integer :: ind_sim, indy,indb,bpsol,indpb



bval=0.
simcredit(1)=1
simb(1)=bval
Do ind_sim=1,Nsim-1
    indy=simindex(ind_sim)
    simy(ind_sim)=endow(indy)
    call gridlookup(bnotsb,bgrid,bval,indb)
    wb=(bgrid(indb+1)-bval)/(bgrid(indb+1)-bgrid(indb))
    If (simcredit(ind_sim)==1) then
        If ((spline==2).and.(soltype==1)) then
            call CubicNkD1(2,bnotsb,bgrid,gb(:,indy),coef1)
            call CubicvalD1(bval,bnotsb,bgrid,gb(:,indy),coef1,bpval)
            simb(ind_sim+1)=bpval
            !call gridlookup(bnotsb,bgrid,bpval,bpsol)
            If (bpval>=0.) then
                aaa=(1./rf)*bpval
            else
                !wbp=(bgrid(bpsol+1)-bpval)/(bgrid(bpsol+1)-bgrid(bpsol))
                !aaa=(wbp*bprice(bpsol,indy)+(1.-wbp)*bprice(bpsol+1,indy))*bpval
                call CubicNkD1(2,bnotsb,bgrid,qb(:,indy),coef2)
                call CubicvalD1(bpval,bnotsb,bgrid,qb(:,indy),coef2,aaa)
            end if
            simqb(ind_sim)=aaa
            simcons(ind_sim)=bval+endow(indy)-aaa
            
            If (bpval>=0.) then
                simprice(ind_sim)=(1/rf)**-1.
            else
                simprice(ind_sim)=(simqb(ind_sim)/bpval)**-1.
            end if
        
          
            simdef(ind_sim)=real(nint(wb*gd(indb,indy)+(1.-wb)*gd(indb+1,indy)))                
        else 
            If (bval<bgrid(Nb)) then
                bvalue=wb*bgrid(indb)+(1.-wb)*bgrid(indb+1)
                simb(ind_sim+1)=wb*gb(indb,indy)+(1.-wb)*gb(indb+1,indy)
                bpval=simb(ind_sim+1)
                call gridlookup(bnotsb,bgrid,bpval,bpsol)
                wbp=(bgrid(bpsol+1)-bpval)/(bgrid(bpsol+1)-bgrid(bpsol))
                simqb(ind_sim)=wbp*qb(bpsol,indy)+(1.-wbp)*qb(bpsol+1,indy)
                simcons(ind_sim)=bval+endow(indy)-simqb(ind_sim)

                If (bpval==0.) then
                    simprice(ind_sim)=(1/rf)**-1.
                else
                    simprice(ind_sim)=(simqb(ind_sim)/bpval)**-1.
                end if
                simdef(ind_sim)=real(nint(wb*gd(indb,indy)+(1.-wb)*gd(indb+1,indy)))
            else
                wb=1.
                indb=Nb
                bvalue=bgrid(indb)
                simb(ind_sim+1)=gb(indb,indy)
                bpval=simb(ind_sim+1)
                simqb(ind_sim)=qb(indb,indy)
                simcons(ind_sim)=bval+endow(indy)-qb(indb,indy)
                    
                If (bpval==0.) then
                    simprice(ind_sim)=(1/rf)**-1.
                else
                    simprice(ind_sim)=(simqb(ind_sim)/bpval)**-1.
                end if
                simdef(ind_sim)=real(nint(gd(indb,indy)))
            end if
        end if
        
        If (simdef(ind_sim)==0.) then
            simcredit(ind_sim+1)=1
        else if (simdef(ind_sim)==1.) then
            simcredit(ind_sim+1)=2
        end if
    else if (simcredit(ind_sim)==2) then
        simb(ind_sim+1)=0.
        simqb(ind_sim)=0.
        simprice(ind_sim)=0.
        If (endow(indy)>ybar) then
            simcons(ind_sim)=ybar
        else
            simcons(ind_sim)=endow(indy)
        end if
        simdef(ind_sim)=0.
        
        aaa=iidseed(ind_sim)
        If (aaa<=theta) then
            simcredit(ind_sim+1)=1
        else
            simcredit(ind_sim+1)=2
        end if  
    end if
    bval=simb(ind_sim+1)
end do
end subroutine simulation


!Subroutine Eulererr(seqnum,simindex,proy,simcons,simb,simcredit,endow,bprice,qb,gb,gc,Dqb,bgrid,simerr,simerrb,averr,maxerr)
!
!Implicit none
!Integer, intent(in) :: seqnum, simindex(Nsim),simcredit(Nsim)
!Real(8), intent(in) :: proy(Ny,Ny),simcons(Nsim),simb(Nsim),endow(Ny)
!Real(8), intent(in) :: gb(Nb,Ny),gc(Nb,Ny),bprice(Nb,Ny),qb(Nb,Ny)
!Real(8), intent(in) :: Dqb(Nb,Ny), bgrid(Nb)
!Real(8), intent(out) :: simerr(seqnum), simerrb(seqnum), averr, maxerr
!Real(8) :: aaa, bbb, ccc, ddd, pry(Ny), bval, bpval, cons(Ny),wb,wbp,expcons,cstar(seqnum)
!Real(8) :: diffqb,bppval,wbpp,consval,coef1(4,bnotsb+1), coef2(4,bnotsb+2),bprval,diffpr
!Integer :: ind_b1, ind_b2, ind_y1, ind_y2, i1, i2, j1, j2, ind_sim,indb,indy,ernum,indbp,countnum,bpsol
!
!simerr=0.
!ernum=0
!bval=0.
!Do ind_sim=1, Nsim-1
!    If ((simcredit(ind_sim)==1).and.(simcredit(ind_sim+1)==1)) then
!        indy=simindex(ind_sim)
!        pry=proy(indy,:)
!        bpval=simb(ind_sim+1)
!        call gridlookup(bnotsb,bgrid,bpval,indbp)
!        wbp=(bgrid(indbp+1)-bpval)/(bgrid(indbp+1)-bgrid(indbp))
!        diffqb=wbp*Dqb(indbp,indy)+(1.-wbp)*Dqb(indbp+1,indy)
!        If (diffqb>0.) then
!            ernum=ernum+1
!            If ((spline==2).and.(soltype==1)) then
!                Do ind_y2=1,Ny
!                    call CubicNkD1(2,bnotsb,bgrid,gb(:,ind_y2),coef1)
!                    call CubicvalD1(bpval,bnotsb,bgrid,gb(:,ind_y2),coef1,bppval)
!                    call gridlookup(bnotsb,bgrid,bppval,bpsol)
!                    wbpp=(bgrid(bpsol+1)-bppval)/(bgrid(bpsol+1)-bgrid(bpsol))
!                    aaa=(wbpp*bprice(bpsol,indy)+(1.-wbpp)*bprice(bpsol+1,indy))*bppval
!                    cons(ind_y2)=bpval+endow(ind_y2)-aaa
!                end do
!            else 
!                If (bpval<bgrid(Nb)) then
!                    Do ind_y2=1,Ny
!                        cons(ind_y2)=wbp*gc(indbp,ind_y2)+(1.-wbp)*gc(indbp+1,ind_y2)
!                    end do
!                else
!                    indbp=Nb
!                    wbp=1.
!                    Do ind_y2=1,Ny
!                        cons(ind_y2)=gc(indbp,ind_y2)
!                    end do
!                end if
!            end if
!            expcons=0.
!            Do ind_y2=1,Ny
!                expcons=expcons+(beta/diffqb)*pry(ind_y2)*cons(ind_y2)**(-risksig)
!            end do
!            cstar(ernum)=expcons**(dble(-1)/risksig)
!            ccc=cstar(ernum)
!            ddd=simcons(ernum)
!            simerr(ernum)=abs(1-(ccc/ddd))
!            simerrb(ernum)=bpval
!        end if
!    end if
!end do
!
!averr=(1./real(ernum))*sum(simerr)
!
!maxerr=maxval(simerr)
!
!        
!end subroutine Eulererr
        
Subroutine distribution(g0,zbmu,bgrid,bprice,gb,gc,gd,Vd,Mval,proy,endow,bgridmu,bpmu,gbmu,gcmu,gdmu,mvmu,g)

implicit none
Integer, intent(in) :: zbmu
Real(8), intent(in) :: proy(Ny,Ny),endow(Ny)
Real(8), intent(in) :: gb(Nb,Ny),gc(Nb,Ny),bprice(Nb,Ny),bgrid(Nb),bgridmu(nbmu)
Real(8), intent(in) :: gd(Nb,Ny), mval(Nb,ny), vd(ny), g0(:,:)
Real(8), intent(out) :: bpmu(nbmu,ny),gbmu(nbmu,ny),gcmu(nbmu,ny),gdmu(nbmu,ny),mvmu(nbmu,ny),g(nbmu+1,ny)
Real(8) :: distance,bval,wgt,g1(nbmu+1,ny),val,bprime,dg(nbmu+1,ny)
Integer :: iteration,ibb,iii(nb),iiii(nbmu), iy,ib,iy1

g1 = g0;
g = 0.0d0


! map the solution on finer grid
do iy = 1,ny; do ib = 1,nbmu

    bval = bgridmu(ib);
    iii = bgrid .le. bval
    ibb = -sum( iii ); ibb = min(ibb, nb-1)

    wgt = ( bgrid(ibb+1)-bval )/( bgrid(ibb+1)-bgrid(ibb) ); wgt = dmax1(wgt,0.0d0); wgt = dmin1(wgt,1.0d0)
    bpmu(ib,iy) = wgt*bprice(ibb,iy) + (1.0d0 - wgt)*bprice(ibb+1,iy)
    gbmu(ib,iy) = wgt*gb(ibb,iy) + (1.0d0 - wgt)*gb(ibb+1,iy)
    gcmu(ib,iy) = wgt*gc(ibb,iy) + (1.0d0 - wgt)*gc(ibb+1,iy)
    mvmu(ib,iy) = wgt*mval(ibb,iy) + (1.0d0 - wgt)*mval(ibb+1,iy)

    if ( mvmu(ib,iy) .le. vd(iy) ) then
        gdmu(ib,iy) = 1.0d0
    end if

end do
end do


distance = 2.0d0; iteration = 0
do while ( distance > eps/1000000.0d0 )

    iteration = iteration + 1
    g = 0.0d0

    !!!!! no default flag !!!!!!!!!!!!!!!!!!!!!!!
    do iy = 1,ny
        do ib = 1,nbmu

            val = g1(ib,iy)

            if (val .gt. 0.0d0) then

                if ( gdmu(ib,iy) .eq. 1.0d0 ) then ! default

                    do iy1 = 1,ny
                        g(nbmu+1,iy1) = g(nbmu+1,iy1) + proy(iy,iy1)*val
                    end do

                else ! repay

                    bprime = gbmu(ib,iy)
                    iiii = bgridmu .le. bprime
                    ibb = -sum( iiii ); ibb = min(ibb, nbmu-1)
                    wgt = ( bgridmu(ibb+1)-bprime )/( bgridmu(ibb+1)-bgridmu(ibb) ); wgt = dmax1(wgt,0.0d0); wgt = dmin1(wgt,1.0d0)

                    do iy1 = 1,ny
                        g(ibb,iy1) = g(ibb,iy1) + proy(iy,iy1)*val*wgt
                        g(ibb+1,iy1) = g(ibb+1,iy1) + proy(iy,iy1)*val*(1.0d0 - wgt)
                    end do

                end if ! default or not

            end if ! if there is mass

        end do
    end do

    !!!!! default flag !!!!!!!!!!!!!!!!!!!!!!!
    g(zbmu,:) = g(zbmu,:) + g(nbmu+1,:)*theta
    g(nbmu+1,:) = g(nbmu+1,:) - g(nbmu+1,:)*theta

    !print*,'mass',sum(g)
    g = g/sum(g)

    dg = abs(g1 - g)
    g1 = g
    distance = maxval(dg)


    if (mod(iteration,100) .eq. 0 ) then
        Write(*, '(a)', Advance='no') 'max(g - gnew) and # of iterations:'
        print*, distance, iteration
    end if


end do

    end subroutine distribution


Subroutine bellmanerror(zbmu,Vd,proy,endow,bgridmu,bpmu,gbmu,gcmu,gdmu,mvmu,g,bellman,averr,maxerr,minerr)

implicit none
Integer, intent(in) :: zbmu
Real(8), intent(in) :: proy(Ny,Ny),endow(Ny),bgridmu(nbmu),vd(:)
Real(8), intent(in) :: bpmu(nbmu,ny),gbmu(nbmu,ny),gcmu(nbmu,ny),gdmu(nbmu,ny),mvmu(nbmu,ny),g(nbmu+1,ny)
Real(8), intent(out) :: bellman(nbmu,ny),averr,maxerr,minerr
Real(8) :: ev(nbmu,ny),cstar(nbmu,ny),bval,wgt,bprime,cval,eval,bellman1(nbmu,ny),util
Integer :: ibb,iiii(nbmu), iy,ib,iy1,iii(nbmu,ny)


ev = 0.0d0
do iy = 1,ny; do iy1 = 1,ny
ev(:,iy) = ev(:,iy) + mvmu(:,iy1)*proy(iy,iy1)
end do
end do
ev = beta*ev 

cstar = 0.0d0
do iy = 1,ny; do ib = 1,nbmu

bprime = gbmu(ib,iy) 
iiii = bgridmu .le. bprime
ibb = -sum( iiii ); ibb = min(ibb, nbmu-1)
wgt = ( bgridmu(ibb+1)-bprime )/( bgridmu(ibb+1)-bgridmu(ibb) ); wgt = dmax1(wgt,0.0d0); wgt = dmin1(wgt,1.0d0)

eval = ev(ibb,iy)*wgt + ev(ibb+1,iy)*(1.0d0-wgt)
cstar(ib,iy) = ( (mvmu(ib,iy) - eval)*(1-risksig) )**( 1/(1-risksig) )

end do
end do

bellman = abs(1-cstar/gcmu)
averr = sum(g(1:nbmu,:)*bellman)

iii = ( g(1:nbmu,:) .gt. 0.0d0 ) .or. ( gdmu .eq. 0.0d0 )
iii = -iii

iii = ( g(1:nbmu,:)*bellman .gt. 0.0d0 ) 
iii = -iii

bellman1 = bellman
bellman1 = iii*bellman1 + (1-iii)*100000.0d0
minerr = minval(bellman1)

bellman1 = bellman
bellman1 = iii*bellman1 + (1-iii)*(-100000.0d0)
maxerr = maxval(bellman1)

print*, 'averr', averr
print*, 'minerr', minerr
print*, 'maxerr', maxerr

end subroutine bellmanerror

End module Inner
    
